--  Copyright  (C) 2006 M. Ben-Ari. See copyright.txt

with Semaphore_Package;
with Ada.Calendar; use Ada.Calendar;
with Ada.Calendar.Formatting;
with Ada.Text_IO;
with Ada.Finalization;

procedure CL is -- for Chandy Lamport
   use Semaphore_Package;

   Start_Time : constant Ada.Calendar.Time :=  Ada.Calendar.Clock;

   type Node_Count is range 0 .. 4; --  up to 4 nodes in the system
   subtype Node_ID is Node_Count range 1 .. Node_Count'Last;

   Marker : constant Integer := -1; -- the marker

   Slow_Down : constant Duration := 1.0; -- small slow-down factor

   type Edge is --  for nodes to know the system topology
      record
         Exists : Boolean := False;
         Last_Message : Integer := 0;
         Recorded_State : Integer := 0;
         Marker_State : Integer := 0;  --  no msg received
      end record;

   --  the whole span of work (can be any)
   subtype Interval is Positive range 1 .. 20;
   --  where the Initiator changes behavior
   Change_Point : constant Interval := 10;

   --  a mutual-exclusion semaphore to protect printing
   Print : constant Binary_Semaphore := Init (1);

   --  a mutual-exclusion semaphore to protect state handling
   Sem : constant Binary_Semaphore := Init (1);

   ------------------- to play with finalization
   type Last_Wishes_T is new Ada.Finalization.Limited_Controlled with
      record
         ID : Node_ID;
      end record;
   procedure Finalize (Last_Wishes : in out Last_Wishes_T) is
      use Ada.Text_IO;
      Passed : Duration;
   begin
      Passed := Ada.Calendar.Clock - Start_Time;
      Put_Line
        (" Node" & Node_ID'Image (Last_Wishes.ID) &
           " terminated at time " & Ada.Calendar.Formatting.Image (Passed));
   end Finalize;
   ----------------------------------

   task type Node is
      ------------------------
      entry Init (ID : Node_ID; N_I, N_O : Node_Count);
      entry Configure (C : Node_ID);
      ------------------------
      entry Message (M : Integer; ID : Node_ID);
   end Node;

   --  the system
   Nodes : array (Node_ID) of Node;
   ----------------------------------

   task body Node is

      --  default initialized
      Incoming : array (Node_ID) of Edge
        with Volatile;

      --  default initialized
      Outgoing : array (Node_ID) of Edge
        with Volatile;

      State_Recorded : Boolean := False
        with Volatile;

      N_In, N_Out : Node_Count := 0;
      Received_ID : Node_ID;
      Received_Data : Integer;

      Current : Node_ID; --  to tell instances of core activity apart

      --------------- to play with finalization
      Last_Wishes : Last_Wishes_T;
      ----------------------------

      procedure Send_Messages (Data : Integer) is
      begin
         for J in Node_ID loop
            if Outgoing (J).Exists then
               Wait (Sem); --  to avoid data races
               if Data >= 0 then
                  Outgoing (J).Last_Message := Data;
               elsif (not State_Recorded) then
                  --  incoming msg is understood as a marker
                  --  state was not recorded yet
                  --  hence it is recorded now
                  Outgoing (J).Recorded_State := Outgoing (J).Last_Message;
               end if;
               Signal (Sem);
               --  node forwards whatever was received
               Nodes (J).Message (Data, Current);
            end if;
         end loop;
      end Send_Messages;

      procedure Record_State is
      begin
         if not State_Recorded then
            Wait (Sem); --  to avoid data races
            for J in Node_ID loop
               if Incoming (J).Exists then
                  Incoming (J).Recorded_State := Incoming (J).Last_Message;
               end if;
            end loop;
            Signal (Sem);
            --  we now send the marker down all outgoing channels
            Send_Messages (Marker);
            --  and record we are in "state recording" mode
            State_Recorded := True;
         end if;
      end Record_State;

      function State_Written return Boolean is
         use Ada.Text_IO;
      begin
         if (not State_Recorded) then
            return False;
         end if;
         --  we are here if we are in "state recording" mode
         for J in Node_ID loop
            if Incoming (J).Exists and (Incoming (J).Marker_State = 0) then
               --  some incoming edge has sent no msg down yet
               --  hence there is work still to do
               return False;
            end if;
         end loop;
         --  we are here if we have recorded all in-flight state
         --  from all edges
         Wait (Print); --  to ensure mutually exclusive print
         Put_Line (" Node" & Node_ID'Image (Current));
         Put_Line ("       Out channels");
         for J in Node_ID loop
            if Outgoing (J).Exists then
               if Outgoing (J).Recorded_State > 0 then
                  --  sent some messages
                  Put_Line
                    ("         " & Node_ID'Image (J) & ": sent 1 .." &
                     Integer'Image (Outgoing (J).Recorded_State));
               end if;
            end if;
         end loop;
         Put_Line ("       In channels");
         for J in Node_ID loop
            if Incoming (J).Exists then
               if Incoming (J).Recorded_State > 0 then
                  --  received some messages
                  Put
                    ("         " & Node_ID'Image (J) &
                     ": received 1 .." &
                     Integer'Image (Incoming (J).Recorded_State));
                  if (Incoming (J).Recorded_State /=
                    Incoming (J).Marker_State)
                  then
                     Put_Line
                       (" - stored " &
                        Integer'Image (Incoming (J).Recorded_State + 1) &
                        " .." & Integer'Image (Incoming (J).Marker_State));
                  else
                     New_Line;
                  end if;
               end if;
            end if;
         end loop;
         Signal (Print);
         return True;
      end State_Written;

      task Core_Activity is
         entry Init;
      end Core_Activity;

      task body Core_Activity is
         --  the same activity for all nodes
         Passed : Duration;
      begin
         --  the node waits for the system topology to be deployed
         accept Init;
         --  the node can now start working
         for J in Interval loop -- the work span
            Send_Messages (J);
            delay Slow_Down;
            --  some specific  nodes are interested in snapshotting
            case Current is
               when 2 | 3 | 4 =>
                  null;
               when 1 =>
                  --  node 1 is the one to send a marker out
                  --  at some point (J) in its life span
                  if J = Change_Point then
                     Record_State;
                     Passed := Ada.Calendar.Clock - Start_Time;
                     Ada.Text_IO.Put_Line
                       ("Node" & Node_ID'Image (Last_Wishes.ID)
                        & " commanded shutdown at time "
                        & Ada.Calendar.Formatting.Image (Passed));
                  end if;
            end case;
         end loop;
         --  control goes here if and when a node had finished work above
         loop
            delay Slow_Down;
            --  this is where Core_Activity should do local work
            --  until the snapshot has completed
            exit when State_Written;
         end loop;
      end Core_Activity;

   begin --  Node
      --  configure node topology
      accept Init (ID : Node_ID; N_I, N_O : Node_Count) do
         Current := ID;
         ----------------------
         Last_Wishes.ID := ID;
         ----------------------
         N_In  := N_I;
         N_Out := N_O;
      end Init;
      ----------------------
      for J in 1 .. N_In loop
         accept Configure (C : Node_ID) do
            Incoming (C).Exists := True;
         end Configure;
      end loop;
      for J in 1 .. N_Out loop
         accept Configure (C : Node_ID) do
            Outgoing (C).Exists := True;
         end Configure;
      end loop;
      ----------------
      --   this call decouples the job of sending messages out
      --   from that of receiving them
      --   (hence the activity of every individual node is strictly concurrent)
      Core_Activity.Init;
      ----------------
      loop
         select
            accept Message (M : Integer; ID : Node_ID) do
               Received_Data := M;
               Received_ID := ID;
            end Message;
            if Received_Data < 0 then
               --  a marker has arrived
               if Incoming (Received_ID).Marker_State = 0 then
                  --  record the last msg received on this edge
                  --  before the marker
                  Incoming (Received_ID).Marker_State :=
                    Incoming (Received_ID).Last_Message;
                  if not State_Recorded then Record_State; end if;
               end if;
            else
               Wait (Sem);
               Incoming (Received_ID).Last_Message := Received_Data;
               Signal (Sem);
            end if;
         or
            terminate;
         end select;
      end loop;
   end Node;

begin -- configure system topology

   Nodes (1).Init (1, 0, 2);
    --  outgoing
   Nodes (1).Configure (2);
   Nodes (1).Configure (3);

   Nodes (2).Init (2, 2, 2);
    --  incoming
   Nodes (2).Configure (1);
   Nodes (2).Configure (3);
    --  outgoing
   Nodes (2).Configure (3);
   Nodes (2).Configure (4);

   Nodes (3).Init (3, 3, 1);
   --  incoming
   Nodes (3).Configure (1);
   Nodes (3).Configure (2);
   Nodes (3).Configure (4);
   --  outgoing
   Nodes (3).Configure (2);

   Nodes (4).Init (4, 1, 1);
    --  incoming
   Nodes (4).Configure (2);
    --  outgoing
   Nodes (4).Configure (3);

   Ada.Text_IO.Put_Line
     ("System topology fully deployed");
   Ada.Text_IO.Put_Line
     ("Per-node workload runs from" &
      Positive'Image (Interval'First) &
      " to" & Positive'Image (Interval'Last));

   --  the main unit now waits for termination
end CL;
